import datetime as dt
import itertools
import math

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    clean_html,
    extract_attributes,
    float_or_none,
    int_or_none,
    merge_dicts,
    parse_iso8601,
    parse_m3u8_attributes,
    parse_resolution,
    urlencode_postdata,
    urljoin,
)
from ..utils.traversal import find_element, traverse_obj


class ZanIE(InfoExtractor):
    IE_NAME = 'zan'
    IE_DESC = 'Z-aN'

    _BASE_URL = 'https://www.zan-live.com'
    _GEO_BYPASS = False
    _GEO_COUNTRIES = ['JP']
    _VALID_URL = r'https?://(www\.)?zan-live\.com/[^/?#]+/live/play/\d+/(?P<id>\d+)'
    _TESTS = [{
        'url': 'https://www.zan-live.com/en/live/play/1797/663',
        'info_dict': {
            'id': '663',
            'ext': 'mp4',
            'title': 'The sample video page',
            'alt_title': 'こちらはサンプル動画の再生テストページとなります。',
            'description': 'md5:12ba331396215fe345b9362c56f3da86',
            'release_date': '20220228',
            'release_timestamp': 1646060400,
            'thumbnail': r're:https?://storage\.zan-live\.com/image/.+\.(?:jpe?g|png)',
        },
    }, {
        'url': 'https://www.zan-live.com/ja/live/play/6910/4268',
        'info_dict': {
            'id': '4268',
            'ext': 'mp4',
            'title': '縁結びのゆかり様1stイベント 結び、結ばれ、桃源郷',
            'alt_title': '縁結びのゆかり様1stイベント 結び、結ばれ、桃源郷 昼公演[アーカイブ]',
            'description': 'md5:f53156a49ce5b45265d07966679ab494',
            'release_date': '20260614',
            'release_timestamp': 1781419500,
            'thumbnail': r're:https?://storage\.zan-live\.com/image/.+\.(?:jpe?g|png)',
        },
    }]

    @staticmethod
    def _fixup_m3u8_formats(formats, m3u8_doc, m3u8_url):
        for stream_inf, media_url in itertools.pairwise(
            map(str.strip, m3u8_doc.splitlines()),
        ):
            if not stream_inf.startswith('#EXT-X-STREAM-INF:') or not media_url:
                continue

            res = traverse_obj(stream_inf, (
                {parse_m3u8_attributes}, 'DISPLAY-NAME', {parse_resolution}))
            if not res:
                continue

            format_url = urljoin(m3u8_url, media_url)
            for fmt in traverse_obj(formats, (
                lambda _, v: v.get('url') == format_url,
            )):
                fmt.update(merge_dicts(fmt, res))

    @staticmethod
    def _get_multiangle_layout(ma_type, ma_number):
        # Z-aN packs all angles into one video frame
        # Square divisions use an n x n grid
        # 6 divisions use a 3 x 3 grid where angle 1 covers a 2 x 2 block
        # https://static.zan-live.com/static/js/live/shaka_multiangle.js

        divisions = int_or_none(ma_type.partition('_')[0]) or 0
        if divisions <= 0:
            return None, None

        unit = math.isqrt(divisions)
        if unit ** 2 == divisions:
            areas = [(i % unit, i // unit, 1, 1) for i in range(divisions)]
        elif divisions == 6:
            unit = 3
            areas = [
                (0, 0, 2, 2),
                *[(2, i, 1, 1) for i in range(2)],
                *[(i, 2, 1, 1) for i in range(3)],
            ]
        else:
            return None, None

        if not (ma_number and 1 <= ma_number <= len(areas)):
            ma_number = len(areas)

        return unit, areas[:ma_number]

    @staticmethod
    def _multiangle_crop(x, y, w, h, unit, margin_y):
        margin_x = margin_y * 9 / 16

        return (
            f'crop='
            f'trunc((iw*{w}/{unit}-{margin_x * 2:g})/2)*2:'
            f'trunc((ih*{h}/{unit}-{margin_y * 2:g})/2)*2:'
            f'trunc(iw*{x}/{unit}+{margin_x:g}):'
            f'trunc(ih*{y}/{unit}+{margin_y:g})'
        )

    def _multiangle_formats(self, formats, ma_type, ma_number, ma_margin):
        if self._configuration_arg('split_angles', ['false'])[0] == 'false':
            self.to_screen(
                'Multi-angle formats are available. Use --extractor-args '
                '"zan:split_angles=true" to extract separate angle formats')
            return formats

        unit, areas = self._get_multiangle_layout(ma_type, ma_number)
        if areas is None:
            self.report_warning(f'Unsupported multiangle type: {ma_type}')
            return formats

        angle_formats = []
        for fmt in formats:
            if fmt.get('vcodec') == 'none':
                continue

            height = traverse_obj(fmt, ('height', {int_or_none}))
            for i, (x, y, w, h) in enumerate(areas, 1):
                angle_formats.append({
                    **fmt,
                    'downloader_options': {
                        'ffmpeg_args_out': [
                            '-vf', self._multiangle_crop(x, y, w, h, unit, ma_margin),
                            '-c:v', 'libx264',
                            '-c:a', 'copy',
                        ],
                    },
                    'format_id': f'{fmt["format_id"]}-angle{i}',
                    'height': int_or_none(
                        (height * h / unit - ma_margin * 2) // 2 * 2) if height else None,
                    'protocol': 'm3u8',
                    'source_preference': -i,
                })

        return angle_formats

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)
        if error_msg := traverse_obj(webpage, (
            {find_element(cls='p-common_message__headline--error')}, {clean_html}, filter,
        )):
            self.raise_geo_restricted(error_msg, countries=self._GEO_COUNTRIES)

        csrf_token = self._html_search_meta('csrf-token', webpage, default=None)
        pct = self._html_search_meta('vod-pct', webpage, default=None)
        token = self._html_search_meta('live-player-token', webpage, default=None)
        if not all((csrf_token, pct, token)):
            self.raise_login_required()

        status = self._download_json(
            f'{self._BASE_URL}/api/live/{video_id}/getLiveStatus', video_id, headers={
                'X-Csrf-Token': csrf_token,
            }, data=urlencode_postdata({
                'pct': pct,
                'token': token,
            }))
        if not traverse_obj(status, ('isSuccess', {bool})):
            raise ExtractorError('Unexpected error')

        result = traverse_obj(status, ('result', {dict}))
        for key, required, error_message in (
            ('isFinished', False, 'This video is no longer available'),
            ('canPlay', True, 'Ticket has expired'),
        ):
            if traverse_obj(result, (key, {bool})) is not required:
                raise ExtractorError(error_message, expected=True)

        is_live = not traverse_obj(result, ('isVod', {bool}))
        release_timestamp = parse_iso8601(self._html_search_meta('open-live-date', webpage))
        srv_time = traverse_obj(status, ('srvTime', {int_or_none}), default=0)

        if is_live and release_timestamp and srv_time < release_timestamp:
            start_time = dt.datetime.fromtimestamp(
                release_timestamp, dt.timezone.utc,
            ).astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')
            self.raise_no_formats(
                f'This livestream is scheduled to start at {start_time}', expected=True)

            return {
                'id': video_id,
                'live_status': 'is_upcoming',
                'release_timestamp': release_timestamp,
            }

        m3u8_url = self._html_search_meta('live-url', webpage, fatal=True)
        m3u8_doc, urlh = self._download_webpage_handle(
            m3u8_url, video_id, note='Downloading m3u8 information')
        m3u8_url = urlh.url

        formats, _ = self._parse_m3u8_formats_and_subtitles(m3u8_doc, m3u8_url, 'mp4')
        self._fixup_m3u8_formats(formats, m3u8_doc, m3u8_url)

        if ma_type := self._html_search_meta('multiangle-type', webpage, default=None):
            ma_number = int_or_none(self._html_search_meta(
                'multiangle-number', webpage, default=None))
            ma_margin = float_or_none(self._html_search_meta(
                'multiangle-margin', webpage, default=None), default=0)
            formats = self._multiangle_formats(formats, ma_type, ma_number, ma_margin)

        detail_url = traverse_obj(webpage, (
            {find_element(cls='linkTxt')},
            {find_element(cls='d-flex align-items-center', html=True)},
            {extract_attributes}, 'href', {urljoin(f'{self._BASE_URL}/')}))
        detail = self._download_webpage(detail_url, video_id, fatal=False) or ''

        return {
            'id': video_id,
            'alt_title': traverse_obj(webpage, (
                {self._og_search_title}, {clean_html}, filter)),
            'formats': formats,
            'is_live': is_live,
            'release_timestamp': release_timestamp,
            'thumbnail': self._og_search_thumbnail(detail),
            **traverse_obj(detail, {
                'title': ({self._og_search_title}, {clean_html}, filter),
                'description': ((
                    {find_element(cls='p-eventinfo__detail')},
                    {find_element(cls='groupDetail')},
                ), {clean_html}, filter, any),
            }),
        }



import functools
import itertools

from .common import InfoExtractor
from ..utils import (
    OnDemandPagedList,
    clean_html,
    filter_dict,
    float_or_none,
    int_or_none,
    parse_duration,
    parse_iso8601,
    parse_qs,
    update_url,
    url_or_none,
)
from ..utils.traversal import (
    require,
    traverse_obj,
    trim_str,
)


class OmnyfmIE(InfoExtractor):
    _VALID_URL = r'https?://omny\.fm/shows/(?P<uploader_id>[\w-]+)/(?P<id>(?!playlists(?:[/?#"\']|$))[\w-]+)(?:/embed)?(?=[?#"\']|$)'
    _EMBED_REGEX = [rf'<iframe[^>]+\bsrc\s*=\s*(["\'])(?P<url>{_VALID_URL}[^"\']*)\1']
    _TESTS = [{
        'url': 'https://omny.fm/shows/sleep-hub/cannabinoids-and-sleep',
        'md5': 'e45ec0ce43da757a0be6ca117ec01bdc',
        'info_dict': {
            'id': 'cannabinoids-and-sleep',
            'ext': 'mp3',
            'title': 'Cannabinoids and Sleep',
            'categories': 'count:1',
            'chapters': [
                {'start_time': 0, 'title': 'Introduction'},
                {'start_time': 138, 'title': 'Theme: Cannabinoids and Sleep'},
                {'start_time': 1487, 'title': 'Clinical Tip'},
                {'start_time': 1635, 'title': 'Pick of the Month'},
                {'start_time': 1795, 'title': 'What\'s Coming Up?'},
            ],
            'description': 'md5:c0fd2d29f3148382d344cfbd012fb00d',
            'duration': 1840.274,
            'episode': 'Episode 48',
            'episode_number': 48,
            'modified_date': r're:\d{8}',
            'modified_timestamp': int,
            'tags': 'count:6',
            'thumbnail': r're:https?://www\.omnycontent\.com/.+',
            'timestamp': 1574013600,
            'upload_date': '20191117',
            'uploader': 'Sleep Talk',
            'uploader_id': 'sleep-hub',
        },
    }, {
        'url': 'https://omny.fm/shows/the-origin-of-things/a-song-of-hope/embed',
        'md5': 'd7600ef33e3f139ff1bb8946f3651b15',
        'info_dict': {
            'id': 'a-song-of-hope',
            'ext': 'mp3',
            'title': 'A song of hope',
            'categories': 'count:3',
            'description': 'md5:f8e710765c341a48cfda8dacd428f56d',
            'duration': 478.955,
            'episode': 'Episode 17',
            'episode_number': 17,
            'modified_date': r're:\d{8}',
            'modified_timestamp': int,
            'season': 'Season 3',
            'season_number': 3,
            'tags': 'count:27',
            'thumbnail': r're:https?://www\.omnycontent\.com/.+',
            'timestamp': 1679445000,
            'upload_date': '20230322',
            'uploader': 'The Origin Of Things',
            'uploader_id': 'the-origin-of-things',
        },
    }]
    _WEBPAGE_TESTS = [{
        'url': 'https://www.asahi.com/special/podcasts/item/?itemid=311a5f48-ad71-4548-b1f2-af5e00747fbc',
        'md5': '4c788bf03323734524a7c0f98d9956ed',
        'info_dict': {
            'id': 'sdgs-271',
            'ext': 'mp3',
            'title': '「どこかのだれかの人生のにおいがする」 SDGsを音声番組で身近に #271',
            'categories': 'count:5',
            'description': 'md5:fa086ecce764d81c51648a82d0fe4850',
            'duration': 1870.524,
            'episode': 'Episode 271',
            'episode_number': 271,
            'modified_date': r're:\d{8}',
            'modified_timestamp': int,
            'season': 'Season 1',
            'season_number': 1,
            'tags': 'count:4',
            'thumbnail': r're:https?://www\.omnycontent\.com/.+',
            'timestamp': 1670266800,
            'upload_date': '20221205',
            'uploader': '朝日新聞ポッドキャスト',
            'uploader_id': 'asahi',
            'webpage_url': 'https://omny.fm/shows/asahi/sdgs-271',
        },
    }]

    def _real_extract(self, url):
        uploader_id, audio_id = self._match_valid_url(url).group('uploader_id', 'id')
        webpage = self._download_webpage(url, audio_id)
        nextjs_data = self._search_nextjs_data(webpage, audio_id)
        clip = traverse_obj(nextjs_data, ('props', 'pageProps', 'clip', {dict}))

        return {
            'id': audio_id,
            'section_start': traverse_obj(url, ({parse_qs}, 't', -1, {parse_duration})),
            'uploader_id': uploader_id,
            'vcodec': 'none',
            **traverse_obj(clip, {
                'title': ('Title', {clean_html}, filter),
                'chapters': ('Chapters', lambda _, v: parse_duration(v['Position']) is not None, {
                    'title': ('Name', {clean_html}, filter),
                    'start_time': ('Position', {parse_duration}),
                }),
                'description': ('Description', {clean_html}, filter),
                'duration': ('DurationSeconds', {float_or_none}),
                'episode_number': ('Episode', {int_or_none}),
                'filesize': ('PublishedAudioSizeInBytes', {int_or_none}),
                'modified_timestamp': ('ModifiedAtUtc', {parse_iso8601}),
                'season_number': ('Season', {int_or_none}),
                'tags': ('Tags', ..., {clean_html}, filter, all, filter),
                'thumbnail': ('ImageUrl', {update_url(query=None)}),
                'timestamp': ('PublishedUtc', {parse_iso8601}),
                'url': ('AudioUrl', {url_or_none}, {require('audio URL')}),
                'webpage_url': ('PublishedUrl', {url_or_none}),
            }),
            **traverse_obj(clip, ('Program', {
                'categories': ('Categories', ..., {clean_html}, filter, all, filter),
                'uploader': ('Name', {clean_html}, filter),
            })),
        }


class OmnyfmPlaylistBaseIE(InfoExtractor):
    _API_BASE = 'https://api.omny.fm'
    _BASE_URL = 'https://omny.fm/shows'
    _PAGE_SIZE = 100

    def _yield_clips(self, clips, uploader_id):
        for audio_id in traverse_obj(clips, (
            'Clips', ..., 'Slug', {str},
        )):
            yield self.url_result(
                f'{self._BASE_URL}/{uploader_id}/{audio_id}', OmnyfmIE)


class OmnyfmPlaylistIE(OmnyfmPlaylistBaseIE):
    _VALID_URL = r'https?://omny\.fm/shows/(?P<uploader_id>[\w-]+)/playlists(?:/(?P<id>[\w-]+))?(?:/embed)?/?(?=[?#"\']|$)'
    _EMBED_REGEX = [fr'<iframe[^>]+\bsrc=(["\'])(?P<url>{_VALID_URL}[^"\']*)\1']
    _TESTS = [{
        'url': 'https://omny.fm/shows/sleep-hub/playlists/sleep-talk',
        'info_dict': {
            'id': 'sleep-talk',
            'title': 'Sleep Talk - Talking all things sleep',
            'description': 'md5:c1d7e5bf32100a432307d2d32c4ab74a',
            'thumbnail': r're:https?://www\.omnycontent\.com/.+',
        },
        'playlist_mincount': 79,
    }, {
        'url': 'https://omny.fm/shows/bayfm-program03/playlists',
        'info_dict': {
            'id': 'bayfm-program03',
        },
        'playlist_count': 4,
    }]
    _WEBPAGE_TESTS = [{
        'url': 'https://www.asahi.com/articles/ASP763WDKP4JDIFI002.html',
        'info_dict': {
            'id': 'podcast',
            'title': 'ニュースの現場から',
            'description': 'md5:ed1f78462ebed09258ca31b1da5ff640',
            'thumbnail': r're:https?://www\.omnycontent\.com/.+',
            'webpage_url': 'https://omny.fm/shows/asahi/playlists/podcast',
        },
        'playlist_mincount': 2517,
    }]

    def _entries(self, uploader_id, playlist_id):
        clip_id = None

        for page in itertools.count(1):
            clips = self._download_json(
                f'{self._API_BASE}/programs/{uploader_id}/playlists/{playlist_id}/clips',
                playlist_id, f'Downloading page {page}', query=filter_dict({
                    'clipId': clip_id,
                    'direction': 'AfterExclusive',
                    'pageSize': self._PAGE_SIZE,
                }))

            yield from self._yield_clips(clips, uploader_id)

            if not clips.get('NextClipsAvailable'):
                break

            clip_id = traverse_obj(clips, ('Clips', -1, 'Id', {str}))
            if not clip_id:
                break

    def _real_extract(self, url):
        uploader_id, playlist_id = self._match_valid_url(url).group('uploader_id', 'id')
        webpage = self._download_webpage(url, playlist_id or uploader_id)
        nextjs_data = self._search_nextjs_data(webpage, playlist_id or uploader_id)
        page_props = traverse_obj(nextjs_data, ('props', 'pageProps', {dict}))

        if not playlist_id:
            entries = [self.url_result(
                f'{self._BASE_URL}/{uploader_id}/playlists/{playlist_id}', OmnyfmPlaylistIE,
            ) for playlist_id in traverse_obj(page_props, (
                'playlistsWithClips', ..., 'playlist', 'Slug', {str},
            ))]

            return self.playlist_result(entries, uploader_id)

        return self.playlist_result(
            self._entries(uploader_id, playlist_id), playlist_id,
            **traverse_obj(page_props, ('playlist', {
                'title': ('Title', {clean_html}, filter),
                'description': ('Description', {clean_html}, filter),
                'thumbnail': ('ArtworkUrl', {update_url(query=None)}),
                'webpage_url': ('EmbedUrl', {url_or_none}, {trim_str(end='/embed')}),
            })))


class OmnyfmShowIE(OmnyfmPlaylistBaseIE):
    _VALID_URL = r'https?://omny\.fm/shows/(?P<id>[\w-]+)/?(?:[?#]|$)'
    _TESTS = [{
        'url': 'https://omny.fm/shows/the-origin-of-things',
        'info_dict': {
            'id': 'the-origin-of-things',
            'title': 'The Origin Of Things',
            'description': 'md5:52b7fba08201d050639c78ea88cc782e',
            'thumbnail': r're:https?://www\.omnycontent\.com/.+',
        },
        'playlist_mincount': 75,
    }]

    def _fetch_page(self, uploader_id, organization_id, program_id, page):
        clips = self._download_json(
            f'{self._API_BASE}/orgs/{organization_id}/programs/{program_id}/clips',
            uploader_id, f'Downloading page {page + 1}', query={
                'cursor': page,
                'pageSize': self._PAGE_SIZE,
            })

        yield from self._yield_clips(clips, uploader_id)

    def _real_extract(self, url):
        uploader_id = self._match_id(url)
        webpage = self._download_webpage(url, uploader_id)
        nextjs_data = self._search_nextjs_data(webpage, uploader_id)
        program = traverse_obj(nextjs_data, ('props', 'pageProps', 'program', {dict}))

        organization_id = traverse_obj(program, (
            'OrganizationId', {str}, {require('organization ID')}))
        program_id = traverse_obj(program, ('Id', {str}, {require('program ID')}))
        entries = OnDemandPagedList(
            functools.partial(self._fetch_page, uploader_id, organization_id, program_id), self._PAGE_SIZE)

        return self.playlist_result(
            entries, uploader_id,
            **traverse_obj(program, {
                'title': ('Name', {clean_html}, filter),
                'description': ('Description', {clean_html}, filter),
                'thumbnail': ('ArtworkUrl', {update_url(query=None)}),
            }))



import itertools
import urllib.parse

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    UnsupportedError,
    clean_html,
    determine_ext,
    float_or_none,
    int_or_none,
    join_nonempty,
    orderedSet,
    parse_iso8601,
    str_or_none,
    url_or_none,
    urljoin,
)
from ..utils.traversal import find_element, traverse_obj


class MxplayerBaseIE(InfoExtractor):
    _BASE_URL = 'https://www.mxplayer.in'
    _GEO_BYPASS = False
    _GEO_COUNTRIES = ['IN']

    def _extract_mxs(self, url, item_id):
        webpage = self._download_webpage(url, item_id)
        if error_msg := traverse_obj(webpage, (
            {find_element(cls='sub-message')}, {clean_html}, filter,
        )):
            self.raise_geo_restricted(error_msg, countries=self._GEO_COUNTRIES)

        return self._search_json(r'window\.__mxs__\s*=', webpage, 'mxs', item_id)


class MxplayerIE(MxplayerBaseIE):
    IE_NAME = 'mxplayer'
    IE_DESC = 'Amazon MX Player'

    _VALID_URL = r'''(?x)
        https?://(?:www\.)?mxplayer\.in/
        (?:movie|shorts|show/[\w-]+/(?!seasons/)[\w-]+)/
        (?P<display_id>[\w-]+)-(?P<id>[0-9a-f]{32})(?:[/?#]|$)
    '''
    _TESTS = [{
        # show, mxplay
        'url': 'https://www.mxplayer.in/show/watch-my-girlfriend-is-an-alien/season-1/episode-1-online-9d2013d31d5835bb8400e3b3c5e7bb72',
        'info_dict': {
            'id': '9d2013d31d5835bb8400e3b3c5e7bb72',
            'ext': 'mp4',
            'title': 'Episode 1',
            'age_limit': 16,
            'cast': 'count:10',
            'creators': 'count:2',
            'description': 'md5:e90dc55a393f557049284eb36efdb773',
            'display_id': 'episode-1-online',
            'episode': 'Episode 1',
            'episode_number': 1,
            'duration': 2451,
            'genres': 'count:8',
            'release_date': '20241201',
            'release_timestamp': 1733077800,
            'season': 'Season 1',
            'season_number': 1,
            'series': 'My Girlfriend Is An Alien',
            'tags': 'count:31',
            'thumbnail': r're:https?://.+',
            'timestamp': 1732435849,
            'upload_date': '20241124',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # show, mxplay, NFSW
        'url': 'https://www.mxplayer.in/show/watch-miya-biwi-aur-murder/season-1/episode-1-online-ca17972f052449de6633129ddd7db90b',
        'info_dict': {
            'id': 'ca17972f052449de6633129ddd7db90b',
            'ext': 'mp4',
            'title': 'Episode 1',
            'age_limit': 18,
            'cast': 'count:10',
            'creators': 'count:1',
            'description': 'md5:bd8e55b81094e55026eb0c4f702c32cc',
            'display_id': 'episode-1-online',
            'episode': 'Episode 1',
            'episode_number': 1,
            'duration': 1348,
            'genres': 'count:4',
            'release_date': '20220630',
            'release_timestamp': 1656613800,
            'season': 'Season 1',
            'season_number': 1,
            'series': 'Miya Biwi Aur Murder',
            'tags': 'count:12',
            'thumbnail': r're:https?://.+',
            'timestamp': 1656581738,
            'upload_date': '20220630',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # shorts, mxplay
        'url': 'https://www.mxplayer.in/shorts/watch-official-trailer-made-in-india-a-titan-story-online-c36b0ad18e56f1da0743ad826fe9e14b',
        'info_dict': {
            'id': 'c36b0ad18e56f1da0743ad826fe9e14b',
            'ext': 'mp4',
            'title': 'Official Trailer | Made In India: A Titan Story',
            'age_limit': 13,
            'description': 'md5:37924586e54d6e86a6e991048aec5442',
            'display_id': 'watch-official-trailer-made-in-india-a-titan-story-online',
            'episode': 'Episode 0',
            'episode_number': 0,
            'duration': 125,
            'genres': 'count:2',
            'release_date': '20260525',
            'release_timestamp': 1779733800,
            'tags': 'count:3',
            'thumbnail': r're:https?://.+',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # movie, mxplay
        'url': 'https://www.mxplayer.in/movie/watch-thiruchitrambalam-movie-online-fbba39c1fdf8b14cffcea05f0635da84',
        'info_dict': {
            'id': 'fbba39c1fdf8b14cffcea05f0635da84',
            'ext': 'mp4',
            'title': 'Thiruchitrambalam',
            'age_limit': 13,
            'cast': 'count:7',
            'creators': 'count:1',
            'description': 'md5:66c24674756099adbecd1e56cfa9d6e6',
            'display_id': 'watch-thiruchitrambalam-movie-online',
            'episode': 'Episode 0',
            'episode_number': 0,
            'duration': 7894,
            'genres': 'count:3',
            'release_date': '20220817',
            'release_timestamp': 1660761000,
            'tags': 'count:8',
            'thumbnail': r're:https?://.+',
            'timestamp': 1759808876,
            'upload_date': '20251007',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # movie, thirdParty
        'url': 'https://www.mxplayer.in/movie/watch-drishyam-2-hindi-movie-online-2c580670b2104d156e2d9a3fd42a413e',
        'info_dict': {
            'id': '2c580670b2104d156e2d9a3fd42a413e',
            'ext': 'mp4',
            'title': 'Drishyam 2 (Hindi)',
            'age_limit': 13,
            'cast': 'count:10',
            'creators': 'count:1',
            'description': 'md5:b4f803532ea357b78ca183d0829eca2a',
            'display_id': 'watch-drishyam-2-hindi-movie-online',
            'episode': 'Episode 0',
            'episode_number': 0,
            'duration': 8399,
            'genres': 'count:3',
            'release_date': '20221117',
            'release_timestamp': 1668709800,
            'tags': 'count:4',
            'thumbnail': r're:https?://.+',
            'timestamp': 1736823124,
            'upload_date': '20250114',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }]

    def _real_extract(self, url):
        display_id, video_id = self._match_valid_url(url).group('display_id', 'id')
        mxs = self._extract_mxs(url, video_id)

        config = traverse_obj(mxs, ('config', {dict}))
        cdn_base = traverse_obj(config, (
            'videoCdnBaseUrl', {url_or_none}, {lambda x: f'{x.rstrip("/")}/'}))
        img_base = traverse_obj(config, (
            'imageBaseUrl', {url_or_none}, {lambda x: f'{x.rstrip("/")}/'}))

        entities = traverse_obj(mxs, ('entities', video_id, {dict}))
        stream = traverse_obj(entities, ('stream', {dict}))
        if traverse_obj(stream, ('drmProtect', {bool})):
            self.report_drm(video_id)

        formats, subtitles = [], {}
        for manifest_url in traverse_obj(stream, ((
            ('thirdParty', ..., {url_or_none}),
            ((None, 'mxplay'), ('hls', 'dash'), 'high', {urljoin(cdn_base)}),
        ), all, {orderedSet})):
            ext = determine_ext(manifest_url)
            if ext == 'm3u8':
                fmts, subs = self._extract_m3u8_formats_and_subtitles(
                    manifest_url, video_id, 'mp4', m3u8_id='hls', fatal=False)
            elif ext == 'mpd':
                fmts, subs = self._extract_mpd_formats_and_subtitles(
                    manifest_url, video_id, mpd_id='dash', fatal=False)
                for fmt in fmts:
                    if tbr := traverse_obj(fmt, ('tbr', {float_or_none})):
                        fmt['tbr'] = tbr / 4.2
            else:
                self.report_warning(f'Unsupported stream type: {ext}')
                continue

            formats.extend(fmts)
            self._merge_subtitles(subs, target=subtitles)
        self._remove_duplicate_formats(formats)

        return {
            'id': video_id,
            'display_id': display_id,
            'formats': formats,
            'subtitles': subtitles,
            **traverse_obj(entities, {
                'title': ('title', {clean_html}, filter),
                'age_limit': ('rating', {int_or_none}),
                'description': ('description', {clean_html}, filter),
                'duration': ('duration', {int_or_none}),
                'episode_number': ('sequence', {int_or_none}),
                'genres': ('genres', ..., {clean_html}, filter, all, filter),
                'release_timestamp': ('releaseDate', {parse_iso8601}),
                'tags': ('tags', ..., 'name', {clean_html}, filter, all, filter),
                'thumbnails': (('imageInfo', 'titleContentImageInfo'), ..., {
                    'id': ('type', {str_or_none}),
                    'height': ('height', {int_or_none}),
                    'url': ('url', {urljoin(img_base)}),
                    'width': ('width', {int_or_none}),
                }),
                'timestamp': ('publishTime', {parse_iso8601}),
                'view_count': ('viewCount', {int_or_none}),
            }),
            **traverse_obj(entities, ('container', {
                'season': ('title', {clean_html}, filter),
                'season_number': ('sequence', {int_or_none}),
                'series': ('container', 'title', {clean_html}, filter),
            })),
            **traverse_obj(entities, ('contributors', {
                'cast': (lambda _, v: v['type'] == 'actor', 'name', {clean_html}, filter, all, filter),
                'creators': (lambda _, v: v['type'] == 'director', 'name', {clean_html}, filter, all, filter),
            })),
        }


class MxplayerSeasonIE(MxplayerBaseIE):
    IE_NAME = 'mxplayer:season'

    _PAGE_SIZE = 10
    _VALID_URL = r'https?://(?:www\.)?mxplayer\.in/show/[\w-]+/seasons/[\w-]+-(?P<id>[0-9a-f]{32})(?:[/?#]|$)'
    _TESTS = [{
        'url': 'https://www.mxplayer.in/show/watch-campus-beats/seasons/season-4-925794ae840c597b3c36f2d8f2f138b8',
        'info_dict': {
            'id': '925794ae840c597b3c36f2d8f2f138b8',
            'title': 'Campus Beats - Season 4',
        },
        'playlist_mincount': 15,
    }]

    def _entries(self, entities, season_id):
        query = None
        api_url = traverse_obj(entities, (
            'tabs', ..., 'api', {urljoin('https://api.mxplayer.in/v1/web/')}, any))

        for page in itertools.count(1):
            season_data = self._download_json(
                api_url, season_id, f'Downloading Page {page}', query=query)

            share_urls = traverse_obj(season_data, (
                'items', ..., 'shareUrl', {urljoin(f'{self._BASE_URL}/')}))
            for share_url in share_urls:
                yield self.url_result(share_url, MxplayerRedirectIE)
            if len(share_urls) < self._PAGE_SIZE:
                break

            query = traverse_obj(season_data, ('next', {urllib.parse.parse_qs}))
            if not query:
                break

    def _real_extract(self, url):
        season_id = self._match_id(url)
        mxs = self._extract_mxs(url, season_id)

        entities = traverse_obj(mxs, ('entities', season_id, {dict}))
        series_title = traverse_obj(entities, ('container', 'title', {clean_html}))
        season_title = traverse_obj(entities, ('title', {clean_html}))

        return self.playlist_result(
            self._entries(entities, season_id),
            season_id, join_nonempty(series_title, season_title, delim=' - '))


class MxplayerShowIE(MxplayerBaseIE):
    IE_NAME = 'mxplayer:show'

    _VALID_URL = r'https?://(?:www\.)?mxplayer\.in/show/[\w-]+-(?P<id>[0-9a-f]{32})(?:[/?#]|$)'
    _TESTS = [{
        'url': 'https://www.mxplayer.in/show/watch-bhaukaal-series-online-775ad3b682fde6a608559a60986b230d',
        'info_dict': {
            'id': '775ad3b682fde6a608559a60986b230d',
            'title': 'Bhaukaal',
        },
        'playlist_mincount': 2,
    }]

    def _entries(self, entities):
        for season_id in traverse_obj(entities, (
            'tabs', ..., 'containers', ..., 'id', {str_or_none},
        )):
            yield self.url_result(
                f'{self._BASE_URL}/detail/season/{season_id}', MxplayerRedirectIE)

    def _real_extract(self, url):
        show_id = self._match_id(url)
        mxs = self._extract_mxs(url, show_id)

        entities = traverse_obj(mxs, ('entities', show_id, {dict}))

        return self.playlist_result(
            self._entries(entities), show_id,
            traverse_obj(entities, ('title', {clean_html}, filter)))


class MxplayerRedirectIE(MxplayerBaseIE):
    IE_NAME = 'mxplayer:redirect'
    IE_DESC = False

    _IE_MAP = {
        'episode': MxplayerIE,
        'movie': MxplayerIE,
        'season': MxplayerSeasonIE,
        'shorts': MxplayerIE,
        'tvshow': MxplayerShowIE,
    }
    _VALID_URL = rf'https?://(?:www\.)?mxplayer\.in/detail/(?P<type>{"|".join(_IE_MAP)})/(?P<id>[0-9a-f]{{32}})(?:[/?#]|$)'
    _TESTS = [{
        # episode
        # https://www.mxplayer.in/show/watch-that-time-i-got-reincarnated-as-a-slime/season-2/megiddo-online-3eda0b3baf27f2892d3fca2fd650fb95
        'url': 'https://www.mxplayer.in/detail/episode/3eda0b3baf27f2892d3fca2fd650fb95',
        'info_dict': {
            'id': '3eda0b3baf27f2892d3fca2fd650fb95',
            'ext': 'mp4',
            'title': 'Megiddo',
            'age_limit': 16,
            'cast': 'count:6',
            'creators': 'count:11',
            'description': 'md5:f4ac8210c2e7bdcaab3b34636cf754b4',
            'display_id': 'megiddo-online',
            'episode': 'Episode 10',
            'episode_number': 10,
            'duration': 1427,
            'genres': 'count:6',
            'release_date': '20210315',
            'release_timestamp': 1615833000,
            'season': 'Season 2',
            'season_number': 2,
            'series': 'That Time I Got Reincarnated As A Slime',
            'tags': 'count:21',
            'thumbnail': r're:https?://.+',
            'timestamp': 1774843170,
            'upload_date': '20260330',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # movie
        # https://www.mxplayer.in/movie/watch-sardar-udham-movie-online-0c586053d7ac563ca911ddfe08cf922f
        'url': 'https://www.mxplayer.in/detail/movie/0c586053d7ac563ca911ddfe08cf922f',
        'info_dict': {
            'id': '0c586053d7ac563ca911ddfe08cf922f',
            'ext': 'mp4',
            'title': 'Sardar Udham',
            'age_limit': 13,
            'cast': 'count:6',
            'creators': 'count:1',
            'description': 'md5:58f0d90f802898aaf328ac2942e878e5',
            'display_id': 'watch-sardar-udham-movie-online',
            'episode': 'Episode 0',
            'episode_number': 0,
            'duration': 9770,
            'genres': 'count:3',
            'release_date': '20211015',
            'release_timestamp': 1634322600,
            'tags': 'count:4',
            'thumbnail': r're:https?://.+',
            'timestamp': 1739952608,
            'upload_date': '20250219',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # shorts
        # https://www.mxplayer.in/shorts/watch-official-trailer-lafangey-sapne-dosti-duniya-online-55c77863e406c7b71f6bab28f7fbbe85
        'url': 'https://www.mxplayer.in/detail/shorts/55c77863e406c7b71f6bab28f7fbbe85',
        'info_dict': {
            'id': '55c77863e406c7b71f6bab28f7fbbe85',
            'ext': 'mp4',
            'title': 'Official Trailer| Lafangey - Sapne, Dosti, Duniya',
            'age_limit': 18,
            'description': 'md5:623d53e4543155dd732e96587711481c',
            'display_id': 'watch-official-trailer-lafangey-sapne-dosti-duniya-online',
            'episode': 'Episode 0',
            'episode_number': 0,
            'duration': 118,
            'genres': 'count:2',
            'release_date': '20250601',
            'release_timestamp': 1748802600,
            'tags': 'count:3',
            'thumbnail': r're:https?://.+',
            'view_count': int,
        },
        'params': {'skip_download': True},
    }, {
        # season
        # https://www.mxplayer.in/show/watch-demon-slayer/seasons/season-3-41a6d7432bea96f2f0d06a625b39d9b1
        'url': 'https://www.mxplayer.in/detail/season/41a6d7432bea96f2f0d06a625b39d9b1',
        'info_dict': {
            'id': '41a6d7432bea96f2f0d06a625b39d9b1',
            'title': 'Demon Slayer - Season 3',
        },
        'playlist_mincount': 11,
    }, {
        # tvshow
        # https://www.mxplayer.in/show/watch-my-hero-academia-series-online-e7c68fb3951e61986af073a719c2ee4f
        'url': 'https://www.mxplayer.in/detail/tvshow/e7c68fb3951e61986af073a719c2ee4f',
        'info_dict': {
            'id': 'e7c68fb3951e61986af073a719c2ee4f',
            'title': 'My Hero Academia',
        },
        'playlist_mincount': 4,
    }]

    def _real_extract(self, url):
        redirect_type, redirect_id = self._match_valid_url(url).group('type', 'id')
        ie = self._IE_MAP[redirect_type]

        detail = self._download_json(
            'https://seo.mxplayer.in/v1/api/seo/get-url-details',
            redirect_id, query={'url': urllib.parse.urlsplit(url).path})

        if redirect_url := traverse_obj(detail, (
            'data', 'redirect', {urljoin(f'{self._BASE_URL}/')},
        )):
            if self.suitable(redirect_url):
                raise UnsupportedError(redirect_url)
            return self.url_result(redirect_url, ie)

        raise ExtractorError('Unable to resolve redirect URL')



# Autogenerated by devscripts/update-version.py

__version__ = '2026.07.04'

RELEASE_GIT_HEAD = '997fa140840a08df3938b40da470c78049fef1f6'

VARIANT = None

UPDATE_HINT = None

CHANNEL = 'stable'

ORIGIN = 'yt-dlp/yt-dlp'

_pkg_version = '2026.07.04'



import collections
import datetime as dt
import functools
import itertools
import json
import math
import time
import urllib.parse
import xml.etree.ElementTree

from .common import InfoExtractor
from ..utils import (
    ExtractorError,
    InAdvancePagedList,
    clean_html,
    extract_attributes,
    filter_dict,
    int_or_none,
    join_nonempty,
    parse_iso8601,
    parse_qs,
    str_or_none,
    update_url_query,
    url_or_none,
    urljoin,
)
from ..utils.traversal import (
    find_element,
    find_elements,
    require,
    traverse_obj,
)


class OpenRecBaseIE(InfoExtractor):
    _API_BASE = 'https://apiv5.mellow-fan.com/api/v5'
    _BASE_URL = 'https://www.mellow-fan.com'
    _HEADERS = {'Referer': f'{_BASE_URL}/'}
    _NETRC_MACHINE = 'mellowfan'
    _PUBLIC_API_BASE = 'https://public.mellow-fan.com/external/api/v5'

    def _perform_login(self, username, password):
        if self._get_cookies(self._BASE_URL).get('access-token'):
            return

        login = self._download_json(
            f'{self._BASE_URL}/apiv5/email/login',
            None, 'Logging in', headers={
                'Content-Type': 'application/json',
            }, data=json.dumps({
                'email': username,
                'password': password,
            }).encode())
        if traverse_obj(login, ('status', {int_or_none})) != 0:
            err_msg = traverse_obj(login, ('message', {clean_html}, filter))
            raise ExtractorError(err_msg or 'Failed to log in', expected=True)

    def _real_initialize(self):
        cookies = self._get_cookies(self._BASE_URL)
        self._api_headers = traverse_obj(cookies, {
            'access-token': ('access_token', 'value', {str}, filter),
            'random': ('random', 'value', {str}, filter),
            'token': ('token', 'value', {str}, filter),
            'uuid': ('uuid', 'value', {str}, filter),
        })

    def _extract_pagestore(self, webpage, video_id):
        start = r'window\.pageStore\s*='

        if store := self._search_regex(
                rf'{start}\s*JSON\.parse\s*\(\s*decodeURIComponent'
                r'\s*\(\s*(?P<q>["\'])(?P<json>.*?)(?P=q)\s*\)\s*\)',
                webpage, 'encoded window pagestore', group='json', default=None,
        ):
            return self._parse_json(store, video_id, transform_source=urllib.parse.unquote)
        return self._search_json(start, webpage, 'window pagestore', video_id)

    def _call_api(self, path, item_id):
        return self._download_json(
            f'{self._API_BASE}/{path}', item_id,
            headers=self._api_headers, expected_status=401)

    def _parse_openrec_metadata(self, page_store, video_id):
        info = traverse_obj(page_store, ('v8', 'movie', {dict}))

        target_members = traverse_obj(info, (
            'targetMembers', ..., 'type', {str}, filter, any))
        needs_subscription = target_members == 'subscription'
        needs_auth = target_members == 'ppv'

        me = self._call_api('users/me', video_id)
        needs_premium = traverse_obj(info, (
            'publicType', {str}, filter)) == 'premium'
        is_premium = traverse_obj(me, (
            'data', 'items', ..., 'is_premium', {bool}, any)) or False

        detail = self._call_api(f'movies/{video_id}/detail', video_id)
        is_member = traverse_obj(detail, (
            'data', 'items', ..., 'membership', 'is_active', {bool}, any)) or False
        has_ppv = traverse_obj(detail, (
            'data', 'items', ..., 'ppv_ticket_products', ..., {dict}, any)) or False

        need = None
        if needs_premium and not is_premium:
            need = 'premium membership'
        elif needs_subscription and not is_member:
            need = 'channel subscription'
        elif needs_auth and not has_ppv:
            need = 'PPV purchase'

        if need:
            self.raise_login_required(
                f'This content requires a {need}', metadata_available=True)

        return info, detail, {
            'id': video_id,
            'availability': self._availability(
                needs_premium=needs_premium,
                needs_subscription=needs_subscription,
                needs_auth=needs_auth,
            ) or 'public',
            'http_headers': self._HEADERS,
            'tags': traverse_obj(page_store, (
                'movieStore', 'tags', ..., {clean_html}, filter, all, filter)),
            **traverse_obj(info, {
                'title': ('title', {clean_html}, filter),
                'cast': ('casts', ..., 'name', {clean_html}, filter, all, filter),
                'categories': ('game', 'title', {clean_html}, filter, all, filter),
                'description': ('introduction', {clean_html}, filter),
                'duration': ('playTime', 'value', {int_or_none(scale=1000)}),
                'thumbnail': (('lThumbnailUrl', 'thumbnailUrl'), {url_or_none}, any),
                'timestamp': ('startedAt', 'time', {int_or_none(scale=1000)}),
                'view_count': ('totalViews', {int_or_none}),
            }),
            **traverse_obj(info, ('channel', 'user', {
                'channel_follower_count': ('followers', {int_or_none}),
                'channel_id': ('id', {str_or_none}),
                'channel': ('name', {clean_html}, filter),
                'channel_is_verified': ('isOfficial', {bool}),
            })),
        }


class OpenRecIE(OpenRecBaseIE):
    IE_NAME = 'mellowfan'
    IE_DESC = 'mellow-fan'

    _VALID_URL = r'https?://(?:www\.)?(?:mellow-fan\.com|openrec\.tv)/(?:m/)?live/(?P<id>[^/?#]+)'
    _TESTS = [{
        'url': 'https://www.openrec.tv/live/e2zwj0mp6ro',
        'info_dict': {
            'id': 'e2zwj0mp6ro',
            'ext': 'mp4',
            'title': '収束',
            'availability': 'public',
            'categories': ['雑談'],
            'channel': 'おおえのたかゆき',
            'channel_follower_count': int,
            'channel_id': 'oekaki',
            'channel_is_verified': True,
            'comment_count': int,
            'description': 'md5:62260f3060b40187282f3213bcd97abd',
            'duration': 14257,
            'live_status': 'was_live',
            'tags': ['雑談'],
            'thumbnail': r're:https?://.+',
            'timestamp': 1685271819,
            'upload_date': '20230528',
            'view_count': int,
        },
        'skip': '404 Not Found',
    }, {
        # SP
        'url': 'https://www.mellow-fan.com/live/2p8vv29438y',
        'info_dict': {
            'id': '2p8vv29438y',
            'ext': 'mp4',
            'title': 'それいけ加藤純一探検隊！ 〜南の孤島で希少生物を探せスペシャル〜 (OPENRECプレミアム会員限定)',
            'availability': 'premium_only',
            'categories': ['雑談'],
            'channel': '加藤 純一',
            'channel_follower_count': int,
            'channel_id': 'junichi_kato_channel',
            'channel_is_verified': True,
            'comment_count': int,
            'description': 'md5:693bc0c838ff6080f9ca00d41dc5e840',
            'duration': 24336,
            'live_status': 'was_live',
            'release_date': '20251008',
            'release_timestamp': 1759892400,
            'thumbnail': r're:https?://.+',
            'timestamp': 1759891802,
            'upload_date': '20251008',
            'view_count': int,
        },
    }, {
        # Members only
        'url': 'https://www.mellow-fan.com/live/kdr7nldqgzj',
        'info_dict': {
            'id': 'kdr7nldqgzj',
            'ext': 'mp4',
            'title': '【ゲーム実況生配信】小森結梨のひきこもりゲーム部屋#20[ゲスト:梅澤めぐ]',
            'availability': 'subscriber_only',
            'categories': ['Human Fall Flat'],
            'channel': 'セカンドショットGAME部',
            'channel_follower_count': int,
            'channel_id': 'secondshot_games',
            'channel_is_verified': True,
            'chapters': [
                {'start_time': 0, 'title': 'バイオハザード RE：2', 'end_time': 443},
                {'start_time': 443, 'title': 'ぷにゃん', 'end_time': 3215},
                {'start_time': 3215, 'title': 'Human Fall Flat', 'end_time': 7163},
            ],
            'comment_count': int,
            'description': 'md5:b77fac6fb0ad86048dc6377ec2c22646',
            'duration': 7163,
            'live_status': 'was_live',
            'release_date': '20260220',
            'release_timestamp': 1771581300,
            'tags': ['声優'],
            'thumbnail': r're:https?://.+',
            'timestamp': 1771581322,
            'upload_date': '20260220',
            'view_count': int,
        },
        'skip': 'Subscribers only',
    }, {
        # PPV
        'url': 'https://www.mellow-fan.com/live/e5rk93xn1zv',
        'info_dict': {
            'id': 'e5rk93xn1zv',
            'ext': 'mp4',
            'title': '三川華月生誕パーティー2026 【ゲスト：幸村恵理/北原沙弥香】',
            'availability': 'needs_auth',
            'categories': ['雑談'],
            'channel': '三川華月の開店！はるちゃん食堂',
            'channel_follower_count': int,
            'channel_id': 'haruna_harusyoku',
            'channel_is_verified': True,
            'description': 'md5:912d0c2d60d343e28300fc18bd491d6c',
            'duration': 6451,
            'live_status': 'was_live',
            'release_date': '20260222',
            'release_timestamp': 1771758000,
            'thumbnail': r're:https?://.+',
            'timestamp': 1771758062,
            'upload_date': '20260222',
            'view_count': int,
        },
        'skip': 'Paid video',
    }]

    @staticmethod
    def _json2xml(subs, started_at):
        def filter_valid(items):
            yield from traverse_obj(items, (
                lambda _, v: clean_html(v['message']) and not v['stamp']))

        total = collections.Counter(
            traverse_obj(subs, (..., 'posted_at', {parse_iso8601})))
        order = collections.defaultdict(int)

        root = xml.etree.ElementTree.Element('packet')
        for i, s in enumerate(filter_valid(subs), 1):
            posted_at = traverse_obj(s, ('posted_at', {parse_iso8601}))
            offset = order[posted_at] / total[posted_at] + 1
            order[posted_at] += 1
            vpos = int_or_none((posted_at - started_at + offset) * 100)

            xml.etree.ElementTree.SubElement(
                root, 'chat', filter_dict({
                    **traverse_obj(s, ('user', {
                        'premium': ('is_premium', {bool}, {lambda x: '1' if x else '0'}),
                        'user_id': ('id', {str_or_none}),
                        'name': ('nickname', {str}, filter),
                    })),
                    'no': str(i),
                    'vpos': str_or_none(vpos),
                    'date': str_or_none(posted_at),
                }),
            ).text = traverse_obj(s, ('message', {clean_html}, filter))
        xml.etree.ElementTree.indent(root, space='  ')

        return xml.etree.ElementTree.tostring(
            root, encoding='utf-8', xml_declaration=True).decode()

    def _get_subtitles(self, duration, started_at, video_id):
        ended_at = started_at + duration
        timestamp = started_at

        subs = []
        for page in itertools.count(1):
            created_at = dt.datetime.fromtimestamp(
                timestamp, dt.timezone(dt.timedelta(hours=9))).strftime('%Y-%m-%dT%H:%M:%S%z')

            chats = self._download_json(
                f'{self._PUBLIC_API_BASE}/movies/{video_id}/chats',
                video_id, f'Downloading chats page {page}', query={
                    'from_created_at': created_at,
                    'is_including_system_message': 'true',
                })
            if not chats:
                break

            subs.extend(chats)
            last_posted_at = traverse_obj(chats, (-1, 'posted_at', {parse_iso8601}))
            if ended_at < last_posted_at:
                break

            timestamp = last_posted_at + 1
            time.sleep(0.1)

        return {
            'chats': [{
                'data': json.dumps(subs, indent=2, ensure_ascii=False),
                'ext': 'json',
            }, {
                'data': self._json2xml(subs, started_at),
                'ext': 'xml',
            }],
        }

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id, expected_status=404)
        page_store = self._extract_pagestore(webpage, video_id)
        if traverse_obj(page_store, ('movieStore', 'notFound', {bool})):
            raise ExtractorError('This video in no longer available', expected=True)

        info, detail, metadata = self._parse_openrec_metadata(page_store, video_id)
        live_status = {
            'ARCHIVE': 'was_live',
            'COMING_UP': 'is_upcoming',
            'LIVE_STREAMING': 'is_live',
            'UPLOADED': 'not_live',
        }.get(info.get('onAirStatus'))

        release_timestamp = traverse_obj(page_store, ('movieStore', 'willStartAt', {parse_iso8601}))
        if live_status == 'is_upcoming':
            if release_timestamp is not None:
                start_time = dt.datetime.fromtimestamp(
                    release_timestamp, dt.timezone.utc,
                ).astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')
                msg = f'This livestream is scheduled to start at {start_time}'
            else:
                msg = 'This livestream has not yet started'

            self.raise_no_formats(msg, expected=True)

            return {
                'id': video_id,
                'live_status': live_status,
                'release_timestamp': release_timestamp,
            }

        duration = metadata['duration']
        started_at = metadata['timestamp']

        chapters = []
        for chapter in traverse_obj(info, (
            'chapters', lambda _, v: int_or_none(v['chapterAt']['time']),
        )):
            chapter_at = traverse_obj(chapter, ('chapterAt', 'time', {int_or_none(scale=1000)}))
            chapters.append({
                'start_time': chapter_at - started_at,
                'title': traverse_obj(chapter, ('title', {clean_html}, filter)),
            })

        formats = []
        is_dvr = live_status == 'is_live' and self.get_param('live_from_start')
        media_keys = ('url_dvr', 'url_dvr_audio') if is_dvr else ('url', 'url_audio')
        for m3u8_url in traverse_obj(detail, (
            'data', 'items', ...,
            ('media', 'subs_trial_media'), media_keys, {url_or_none},
        )):
            fmts = self._extract_m3u8_formats(
                m3u8_url, video_id, 'mp4', headers=self._HEADERS)
            for fmt in fmts:
                if is_dvr:
                    fmt.setdefault('downloader_options', {}).update({'ffmpeg_args': ['-live_start_index', '0']})
                    fmt['is_from_start'] = True
            formats.extend(fmts)

        return {
            'chapters': chapters or None,
            'comment_count': traverse_obj(page_store, (
                'commentStore', 'commentCount', {int_or_none})),
            'formats': formats,
            'live_status': live_status,
            'release_timestamp': release_timestamp,
            'subtitles': self.extract_subtitles(duration, started_at, video_id),
            **metadata,
        }


class OpenRecCaptureIE(OpenRecBaseIE):
    IE_NAME = 'mellowfan:capture'

    _VALID_URL = r'https?://(?:www\.)?(?:mellow-fan\.com|openrec\.tv)/(?:m/)?capture/(?P<id>[^/?#]+)'
    _TESTS = [{
        'url': 'https://www.mellow-fan.com/capture/l2q00vxl8q8',
        'info_dict': {
            'id': 'l2q00vxl8q8',
            'ext': 'mp4',
            'title': '？？？？',
            'channel': '布団ちゃん',
            'channel_id': 'indegnasen',
            'duration': 89,
            'thumbnail': r're:https?://.+',
            'timestamp': 1637589871,
            'upload_date': '20211122',
        },
    }, {
        'url': 'https://www.mellow-fan.com/capture/9pdz9334vng',
        'info_dict': {
            'id': '9pdz9334vng',
            'ext': 'mp4',
            'title': 'オプレの現実',
            'channel': 'ゆゆうた&みゃこの泥沼バラエティ',
            'channel_id': 'doronuma-variety',
            'duration': 64,
            'thumbnail': r're:https?://.+',
            'timestamp': 1677585253,
            'upload_date': '20230228',
        },
    }]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)
        page_store = self._extract_pagestore(webpage, video_id)

        capture = page_store['capture']
        m3u8_url = traverse_obj(capture, ('source', {url_or_none}))

        return {
            'id': video_id,
            'formats': self._extract_m3u8_formats(
                m3u8_url, video_id, 'mp4', headers=self._HEADERS),
            'http_headers': self._HEADERS,
            **traverse_obj(page_store, ('movie', 'channel', {
                'channel': ('name', {clean_html}, filter),
                'channel_id': ('id', {str}),
            })),
            **traverse_obj(capture, {
                'title': ('title', {clean_html}, filter),
                'duration': ({lambda x: int_or_none(x['endTime']) - int_or_none(x['startTime'])}),
                'thumbnail': ('thumbnailUrl', {url_or_none}),
                'timestamp': ('publishedAt', {parse_iso8601}),
            }),
        }


class OpenRecMovieIE(OpenRecBaseIE):
    IE_NAME = 'mellowfan:movie'

    _VALID_URL = r'https?://(?:www\.)?(?:mellow-fan\.com|openrec\.tv)/(?:m/)?movie/(?P<id>[^/?#]+)'
    _TESTS = [{
        'url': 'https://www.mellow-fan.com/movie/e5rk9k4o6zv',
        'info_dict': {
            'id': 'e5rk9k4o6zv',
            'ext': 'mp4',
            'title': 'みゃこRaMuの企画会議#3 ～2人のやってみたいこと～',
            'availability': 'public',
            'categories': ['雑談'],
            'channel': 'みゃことRaMuの何して遊ぶ？',
            'channel_follower_count': int,
            'channel_id': 'myakoramu',
            'channel_is_verified': True,
            'description': 'md5:90924fd73356ebd574bec8d761d4fa62',
            'duration': 771,
            'tags': ['雑談'],
            'thumbnail': r're:https?://.+',
            'view_count': int,
        },
    }, {
        # Members only
        'url': 'https://www.mellow-fan.com/movie/n9ze6q3eo84',
        'info_dict': {
            'id': 'n9ze6q3eo84',
            'ext': 'mp4',
            'title': '大西亜玖璃・高尾奏音のあぐのんる～むらぼ♪第123回傑作選vol.3【高画質・完全版】',
            'availability': 'subscriber_only',
            'categories': ['雑談'],
            'channel': 'セカンドショットGAME部',
            'channel_follower_count': int,
            'channel_id': 'secondshot_games',
            'channel_is_verified': True,
            'description': 'md5:b7ab5ddd71ba5edc1141cb427af7a9c6',
            'duration': 1800,
            'tags': ['声優'],
            'thumbnail': r're:https?://.+',
            'view_count': int,
        },
        'skip': 'Subscribers only',
    }, {
        # PPV
        'url': 'https://www.mellow-fan.com/movie/em8xvd4ljr2',
        'info_dict': {
            'id': 'em8xvd4ljr2',
            'ext': 'mp4',
            'title': '【PPV購入特典映像】KAWAII LAB. SESSION in OKINAWA',
            'availability': 'needs_auth',
            'categories': ['ミュージック'],
            'channel': 'KAWAII LAB.チャンネル',
            'channel_follower_count': int,
            'channel_id': 'KAWAIILAB',
            'channel_is_verified': True,
            'description': 'md5:e9f67d8648d3cbe35df1dd689afc29a8',
            'duration': 2002,
            'tags': ['アイドル'],
            'thumbnail': r're:https?://.+',
            'view_count': int,
        },
        'skip': 'Paid video',
    }]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage = self._download_webpage(url, video_id)
        page_store = self._extract_pagestore(webpage, video_id)
        _, detail, metadata = self._parse_openrec_metadata(page_store, video_id)

        formats = []
        for m3u8_url in traverse_obj(detail, (
            'data', 'items', ..., 'media',
            ('url', 'url_audio'), {url_or_none},
        )):
            formats.extend(self._extract_m3u8_formats(
                m3u8_url, video_id, 'mp4', headers=self._HEADERS))

        return {
            'formats': formats,
            **metadata,
        }


class OpenRecPlaylistIE(OpenRecBaseIE):
    IE_NAME = 'mellowfan:playlist'

    _VALID_URL = r'https?://(?:www\.)?(?:mellow-fan\.com|openrec\.tv)/(?:m/)?user/[^/?#]+/playlist/(?P<id>[^/?#]+)'
    _TESTS = [{
        # live
        'url': 'https://www.mellow-fan.com/user/DbD_BPF/playlist/j59svruhtua2z8t',
        'info_dict': {
            'id': 'j59svruhtua2z8t',
            'title': 'BPFのおすすめ',
        },
        'playlist_mincount': 10,
    }, {
        # capture
        'url': 'https://www.mellow-fan.com/user/sagara_mayu/playlist/xngNMzv71yLjGdW',
        'info_dict': {
            'id': 'xngNMzv71yLjGdW',
            'title': '相良茉優のFAN!FUN!FACTORY!キャプチャ',
        },
        'playlist_mincount': 4,
    }, {
        # movie
        'url': 'https://www.mellow-fan.com/user/oreranohonoka/playlist/sficoshvi9dgkqh',
        'info_dict': {
            'id': 'sficoshvi9dgkqh',
            'title': 'SP（スペシャル）映像：（短編映像）おにいたむプランだけが視聴できます',
        },
        'playlist_mincount': 111,
    }]

    def _entries(self, items):
        for movie in traverse_obj(items, (
            'playlist_movies', ..., 'movie', 'id', {str_or_none},
        )):
            is_live = traverse_obj(movie, ('movie', 'is_live', {bool}))
            path, ie = ('live', OpenRecIE) if is_live else ('movie', OpenRecMovieIE)
            movie_id = movie['movie']['id']

            yield self.url_result(f'{self._BASE_URL}/{path}/{movie_id}', ie)

        for capture in traverse_obj(items, (
            'playlist_captures', ...,
            'capture_relation', 'capture', 'id', {str_or_none},
        )):
            capture_id = capture['capture_relation']['capture']['id']

            yield self.url_result(
                f'{self._BASE_URL}/capture/{capture_id}', OpenRecCaptureIE)

    def _real_extract(self, url):
        playlist_id = self._match_id(url)
        me = self._call_api(f'users/me/playlists/{playlist_id}', playlist_id)

        if items := traverse_obj(me, (
            'data', 'items', ..., {dict}, any,
        )):
            return self.playlist_result(
                self._entries(items), playlist_id,
                traverse_obj(items, ('title', {clean_html}, filter)))

        webpage = self._download_webpage(url, playlist_id)
        playlist_title = traverse_obj(webpage, (
            {find_element(cls='sc-1ak77bz-10')}, {clean_html}, filter))

        return self.playlist_from_matches(traverse_obj(webpage, (
            {find_elements(cls='sc-vc0xhn-0')}, ...,
            {find_element(cls='sc-1ddd11y-0', html=True)},
            {extract_attributes}, 'href', {str},
        )), playlist_id, playlist_title, getter=urljoin(f'{self._BASE_URL}/'))


class OpenRecChannelIE(OpenRecBaseIE):
    IE_NAME = 'mellowfan:channel'

    _PAGE_SIZE = 40
    _VALID_URL = r'https?://(?:www\.)?(?:mellow-fan\.com|openrec\.tv)/(?:m/)?user/(?P<id>[^/?#]+)$'
    _TESTS = [{
        'url': 'https://www.mellow-fan.com/user/OPENRECPARK',
        'info_dict': {
            'id': 'OPENRECPARK',
            'title': 'OPENREC PARK',
        },
        'playlist_mincount': 40,
    }]

    def _fetch_page(self, channel_id, page):
        page += 1
        search_movies = self._download_json(
            f'{self._PUBLIC_API_BASE}/search-movies', channel_id,
            f'Downloading page {page}', query={
                'channel_ids': channel_id,
                'include_live': 'true',
                'include_upload': 'true',
                'onair_status': '2',
                'include_deleted': 'true',
                'sort': 'published_at',
                'page': str(page),
            })

        for movie in traverse_obj(search_movies, (
            lambda _, v: str_or_none(v['movie_type']) and str_or_none(v['id']),
        )):
            path, ie = ('live', OpenRecIE) if movie['movie_type'] == '1' else ('movie', OpenRecMovieIE)

            yield self.url_result(f'{self._BASE_URL}/{path}/{movie["id"]}', ie)

    def _real_extract(self, url):
        channel_id = self._match_id(url)
        webpage = self._download_webpage(url, channel_id)
        page_store = self._extract_pagestore(webpage, channel_id)

        channel = traverse_obj(page_store, ('state', '_channel', {dict}))
        movie_count = traverse_obj(channel, ('movieCount', {int_or_none}))

        return self.playlist_result(InAdvancePagedList(
            functools.partial(self._fetch_page, channel_id),
            math.ceil(movie_count / self._PAGE_SIZE), self._PAGE_SIZE,
        ), channel_id, traverse_obj(channel, ('user', 'name', {clean_html}, filter)))


class OpenRecChannelSearchIE(OpenRecBaseIE):
    IE_NAME = 'mellowfan:channel:search'

    _VALID_URL = r'https?://(?:www\.)?(?:mellow-fan\.com|openrec\.tv)/(?:m/)?user/(?P<id>[^/?#]+)/search(?:/(?P<type>capture|movie))?(?:[/?#]|$)'
    _TESTS = [{
        'url': 'https://www.mellow-fan.com/user/indegnasen/search?search_query=%E3%82%B9%E3%82%A4%E3%82%AB',
        'info_dict': {
            'id': 'indegnasen',
            'title': 'indegnasen:スイカ',
        },
        'playlist_count': 2,
    }, {
        'url': 'https://www.mellow-fan.com/user/ofurekodesu/search/movie?search_query=%E3%82%A2%E3%83%AA%E3%82%AA%E5%85%AB%E5%B0%BE',
        'info_dict': {
            'id': 'ofurekodesu',
            'title': 'ofurekodesu:アリオ八尾:movie',
        },
        'playlist_mincount': 31,
    }, {
        'url': 'https://www.mellow-fan.com/user/DbD_BPF/search/capture?search_query=%E3%81%82%E3%81%A3%E3%81%95%E3%82%8A%E3%81%97%E3%82%87%E3%81%93',
        'info_dict': {
            'id': 'DbD_BPF',
            'title': 'DbD_BPF:あっさりしょこ:capture',
        },
        'playlist_mincount': 10,
    }]

    def _entries(self, channel_id, search_type, search_query):
        api_url = f'{self._PUBLIC_API_BASE}/search-{search_type}s'
        type_map = {
            'capture': OpenRecCaptureIE,
            'live': OpenRecIE,
            'movie': OpenRecMovieIE,
        }

        for page in itertools.count(1):
            search_items = self._download_json(
                api_url, channel_id, f'Downloading page {page}', query={
                    'channel_ids': channel_id,
                    'page': page,
                    'search_query': search_query,
                })
            if not search_items:
                break

            for item in search_items:
                item_type = 'live' if search_type == 'movie' and traverse_obj(item, ('is_live', {bool})) else search_type
                item_id = traverse_obj(item, ((None, 'capture'), 'id', {str_or_none}, any))

                yield self.url_result(
                    f'{self._BASE_URL}/{item_type}/{item_id}', type_map[item_type])

    def _real_extract(self, url):
        channel_id, search_type = self._match_valid_url(url).group('id', 'type')
        search_query = traverse_obj(url, (
            {parse_qs}, 'search_query', -1, {str}, filter,
            {require('search query', expected=True)}))

        if not search_type:
            entries = []
            for search_type in ('capture', 'movie'):
                search_url = update_url_query(
                    f'{self._BASE_URL}/user/{channel_id}/search/{search_type}', {'search_query': search_query})
                entries.append(self.url_result(search_url, OpenRecChannelSearchIE))

            return self.playlist_result(
                entries, channel_id, join_nonempty(channel_id, search_query, delim=':'))

        return self.playlist_result(
            self._entries(channel_id, search_type, search_query),
            channel_id, join_nonempty(channel_id, search_query, search_type, delim=':'))
